# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

import sys
import re
import os
import shutil

# pip install GitPython
# The git executable must be specified in one of the following ways:
#     - be included in your $PATH
#     - be set via $GIT_PYTHON_GIT_EXECUTABLE
#     - explicitly set via git.refresh()

# we need elevate **only** with --install-exe and --install-tools
# but elevate prevents from loading git, because git is not installed for the administrator
#         import git


import argparse
import datetime
import getpass

# this file is run by tox.
# All installation commands are executed using {toxinidir} (the directory where tox.ini resides) as CWD # noqa
# to test without tox, run: 'cd .. ; python piwiPre/pwpPatcher.py  '                                    # noqa
# the following line is mandatory to execute from shell
# tox does it automatically, so not necessary if running only from tox.
sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))

from piwiPre.pwpParser import build_official_rst
from piwiPre.pwpImageEmbedder import PwpEmbedder


class PwpPatcher:
    def __init__(self, arguments=None, standalone=True):
        if not standalone:
            return  # pragma: no cover
        import git
        self.parser = argparse.ArgumentParser(description='manages tags and modifies accordingly files that want it ')
        self.parser.add_argument('--Major', '-M', help="Promotes as Major release i.e. X++.",
                                 action='store_true')
        self.parser.add_argument('--minor', '-m', help="Promotes as minor version ie x.Y++",
                                 action='store_true')
        self.parser.add_argument('--patch', '-p', help="Increments patch version ie x.y.Z++",
                                 action='store_true')
        self.parser.add_argument('--commit', '-c', help="Commits, implicit if -M, -m, -p",
                                 action='store_true')
        self.parser.add_argument('--set-tag', help="Sets the tag. ***NOT*** implicit with -M, -m, -p",
                                 action='store_true')

        self.parser.add_argument('--full-help', help="more explanations",
                                 action='store_true')
        self.parser.add_argument('--print', help="prints the last tag spec and exits",
                                 action='store_true')

        self.parser.add_argument('--download', '-d', help="patches download.rst",
                                 action='store_true')
        self.parser.add_argument('--autotest', help="autotest, no real action",
                                 action='store_true')
        self.parser.add_argument('--push', help="push to remote git server",
                                 action='store_true')
        self.parser.add_argument('--sphinx', help="runs sphinx to update the doc",
                                 action='store_true')

        self.parser.add_argument('--exe', help="builds the exe with PyInstaller",
                                 action='store_true')
        self.parser.add_argument('--pypi', help="uploads package to Pypi",
                                 action='store_true')
        self.parser.add_argument('--gitlab', help="uploads exe to gitlab",
                                 action='store_true')

        self.repo = git.Repo('.')

        self.origin = self.repo.remote("origin")

        tags = sorted(self.repo.tags, key=lambda t: t.commit.committed_datetime)
        if len(tags) > 0:
            latest_tag = tags[-1]
            tag_version = str(latest_tag)
        else:  # pragma: no cover
            # latest_tag = None
            tag_version = "0.0.0"

        full_tag = re.sub(r"[^\d.]*", '', tag_version)
        m = re.match(r"(\d+)\.*(\d*)\.*(\d*)", full_tag)
        self.release = int(m.group(1)) if m.group(1) else 0
        self.version = int(m.group(2)) if m.group(2) else 0
        self.patch = int(m.group(3)) if m.group(3) else 0
        self.spec = full_tag
        # if latest_tag:
        #     self.date = latest_tag.commit.committed_datetime.strftime("%m/%d/%Y %H:%M:%S")  # noqa
        # else:  # pragma: no cover
        #     self.date = datetime.datetime.now()
        self.date = datetime.datetime.now()
        self.update_spec()
        self.args = self.parser.parse_args() if arguments is None else self.parser.parse_args(arguments)

    def update_spec(self):
        self.spec = str(self.release)
        if self.version != 0 or self.patch != 0:
            self.spec += '.' + str(self.version)
            if self.patch != 0:
                self.spec += '.' + str(self.patch)

    def manage_tags(self):

        if self.args.patch:
            self.patch += 1

        if self.args.minor:
            self.version += 1
            self.patch = 0

        if self.args.Major:
            self.release += 1
            self.version = 0
            self.patch = 0

        if self.args.Major or self.args.minor or self.args.patch:
            n = datetime.datetime.now()
            self.date = n.strftime("%m/%d/%Y %H:%M:%S")
            self.update_spec()

            print(f"msg     new tag '{self.spec}'")
            return True
        return False

    @staticmethod
    def open(file_name, mode):
        try:
            ins = open(file_name, mode, encoding="utf-8")
        except OSError as error:
            print(f"FATAL ERROR: Cannot open('{file_name}',{mode}) : {str(error)}, aborting")
            exit(-1)
        return ins

    @staticmethod
    def rename(old, new):
        try:
            os.rename(old, new)
        except OSError as error:
            print(f"FATAL ERROR: Cannot rename('{old}','{new}') : {str(error)}, aborting")
            exit(-1)
        return

    def update_file(self, src: str, dst: str, dico: dict, duplicate=False):

        new_file = dst + '.new'
        old_file = dst + '.bak'

        ins = self.open(src, "r")
        content = ins.readlines()
        ins.close()

        outs = self.open(new_file, "w")
        useful = False
        previous = []
        for line in content:
            modified = line
            for before, after in dico.items():
                modified = re.sub(before, after, modified)
            if modified not in previous:
                previous.append(modified)
            else:
                # we have already done that modification once, we do not do it twice
                modified = line
            outs.writelines(modified)

            if modified != line:
                useful = True
                if duplicate:
                    outs.writelines(line)

        outs.close()

        if not useful and src == dst:  # pragma: no cover
            os.remove(new_file)
            print(f"msg     file '{dst}' does not need patch")
            return
        else:   # pragma: no cover
            if os.path.isfile(old_file):
                os.remove(old_file)

            if src == dst:
                self.rename(src, old_file)
                self.rename(new_file, dst)
            else:
                shutil.copy2(src, old_file)
                shutil.copy2(new_file, dst)

            print(f"msg     file '{dst}' patched")

    @staticmethod
    def update_logo(source, dst):
        if os.path.isfile(dst) and os.path.getmtime(dst) > os.path.getmtime(source):
            print(f"msg     file '{dst}' is older than source '{source}': update useless")  # pragma: no cover
            return

        embedder = PwpEmbedder()
        embedder.resize_and_convert(source, dst, 50, 29)

    def update_from_source(self, source, dst):
        if os.path.getmtime(dst) > os.path.getmtime(source):
            print(f"msg     file '{dst}' is older than source '{source}': patch useless")  # pragma: no cover
            return                                                                              # pragma: no cover

        with self.open(source, "r") as ins:
            content = ins.readlines()

        new_file = dst + '.new'                                                         # pragma: no cover
        old_file = dst + '.bak'

        outs = self.open(new_file, "w")
        do_add = True
        with self.open(dst, "r") as ins:
            for line in ins:
                if do_add:
                    outs.write(line)

                    if re.match(r' *def print\(.*\):\n', line):
                        for cl in content:
                            patched = re.sub('"', '\\"', cl[:-1])
                            outs.write(f'        print("{patched}")\n') # add \t\t\t # noqa ???
                        outs.write('    # End of patched text\n')
                        do_add = False
                else:
                    if re.match(r' +# End of patched text\n', line):
                        do_add = True
        outs.close()

        if os.path.isfile(old_file):
            os.remove(old_file)
        shutil.copy2(source, old_file)
        self.rename(new_file, dst)
        print(f"msg     file '{dst}' patched")

    def run(self):
        if self.args.autotest:
            self.args.full_help = True

        if self.args.print:
            print(self.spec)
            return

        print(f"msg     initial tag '{self.spec}' at '{self.date}'")

        if self.args.full_help:
            print("if -M, -m or -p, updates the current tag")
            print("verifies that the current tag is writen in piwiPre/pwpVersion.py")
            print("verifies that piwiPre/pwpLicence.py is up to date vs source")
            print("if ")
            print("   - the new tag is a Major or a minor (but *not* a patch), then the corresponding exe and module")
            print("     will be published, assuming here that they have been tested OK)")
            print("     therefore download.rst and téléchargement.rst are also modified")
            print("   - or -d")
            print("then download.rst and téléchargement.rst are updated")

            print("if --commit, commits ALL modified files")
            print("     commit, is implicit if set-tag, -M, -m or -p")

            print("if --set-tag, sets the tag locally (and remotely if --push)")
            print("else, do not set the tag, which allows to patch source files before setting the tag")
            print("if a new tag puts the tag LOCALLY on git, but *not* remotely on GitLab")

        if self.args.push:
            self.args.set_tag = True

        if self.args.Major or self.args.minor or self.args.patch or self.args.set_tag:
            self.args.commit = True

        new_tag = self.manage_tags()

        if self.args.autotest:
            self.args.Major = True
            self.manage_tags()
            self.args.Major = False
            self.args.minor = True
            self.manage_tags()
            self.args.minor = False
            self.args.patch = True
            new_tag = self.manage_tags()

        if self.args.autotest:
            shutil.copy2('piwiPre/pwpLicence.py', 'tests/results/pwpLicence.py')

        self.update_logo("pwpLogo.png",
                         "piwiPre/pwpLogoSmall.py" if not self.args.autotest else "tests/results/pwpLogoSmall.py",
                         )
        self.update_from_source('LICENCE',
                                'piwiPre/pwpLicence.py' if not self.args.autotest else 'tests/results/pwpLicence.py')

        url_base = 'https://gitlab.com/api/v4/projects/48915444/packages/generic/piwiPre'

        self.update_file('piwiPre/pwpVersion.py',
                         'piwiPre/pwpVersion.py' if not self.args.autotest else 'tests/results/pwpVersion.py',
                         {r"help = '([^']*)'": f"help = '{self.spec} at {self.date}'",
                          r"spec = '([^']*)'": f"spec = '{self.spec}'",
                          r"release = (\d*)": f"release = {self.release}",
                          r"version = (\d*)": f"version = {self.version}",
                          r"patch = (\d*)": f"patch = {self.patch}",
                          r"date = '([^']*)'": f"date = '{self.date}'",
                          r"url = '([^']*)'": f"url = '{url_base}/{self.spec}/piwiPre-{self.spec}.exe'",
                          r"installer = '([^']*)'": f"url = '{url_base}/{self.spec}/piwiPreInstaller-{self.spec}.exe'"
                          })

        # - Windows one-file exe on gitlab artifacts:
        # `piwiPre-1.8.3.exe
        # <https://gitlab.com/api/v4/projects/22464405/packages/generic/piwiPre/1.8.3/piwiPre-1.8.3.exe>`_

        prologue_en = '- Windows one-file exe on gitlab artifacts:'
        prologue2_en = '- Windows installer on gitlab artifacts:'
        prologue_fr = "- l'exécutable Windows sur le serveur d'artefacts gitlab:"
        prologue2_fr = "- L'installateur Windows sur le serveur d'artefacts gitlab:"

        if self.args.download or (new_tag and self.patch == 0) or self.args.autotest:
            self.update_file('source/download.rst',
                             'source/download.rst' if not self.args.autotest else 'tests/results/download.rst',
                             {
                                 r".*Windows one-file exe.*": f"{prologue_en} `piwiPre-{self.spec}.exe " +
                                                              f"<{url_base}/{self.spec}/piwiPre-{self.spec}.exe>`_",
                                 r".*Windows instal.*": f"{prologue2_en} `pwpInstaller-{self.spec}.exe " +  # noqa
                                                        f"<{url_base}/{self.spec}/pwpInstaller-{self.spec}.exe>`_",
                             },
                             duplicate=True)
            self.update_file('source/fr/téléchargement.rst',
                             'source/fr/téléchargement.rst' if not self.args.autotest else
                             'tests/results/téléchargement.rst',
                             {
                                 r".*exécutable Windows.*": f"{prologue_fr} `piwiPre-{self.spec}.exe " +
                                                            f"<{url_base}/{self.spec}/piwiPre-{self.spec}.exe>`_",
                                 r".*lateur Wind.*": f"{prologue2_fr} `piwiPreInstaller-{self.spec}.exe " +  # noqa
                                                     f"<{url_base}/{self.spec}/piwiPreInstaller-{self.spec}.exe>`_"
                             },
                             duplicate=True)

        self.update_file('version.txt',
                         'version.txt' if not self.args.autotest else 'tests/results/version.txt',
                         {r"(.+)":  self.spec})
        build_official_rst(self.args.autotest)

        if self.args.autotest:
            print(f"msg     final tag '{self.spec}' at '{self.date}'")
            return

        if self.args.sphinx:
            res = os.system("python -m sphinx source public/html -W -b html")
            print(f"msg     DONE: sphinx -> {res}")

        if self.args.commit:
            commit = self.repo.git.commit('-a', '-m', f'set tag {self.spec}', '--date', self.date)
            print(f"msg     DONE: git commit '{str(commit)}' ")

        if self.args.set_tag:
            self.repo.git.tag(self.spec)
            print(f"msg     DONE: git tag '{self.spec}'")

        if self.args.push and self.origin:
            # res = self.repo.git.push("--set-upstream", self.origin, self.repo.head.ref)
            res = self.repo.git.push('origin', self.spec)
            print(f"msg     DONE: git push {res}")

        if self.args.exe:
            res = os.system("python -m PyInstaller -y piwiPre/__main__.py --onefile --name piwiPre")  # noqa
            print(f"msg     DONE: build exe piwiPre -> {res}")
            res = os.system("python -m PyInstaller -y piwiPre/pwpInstaller.py --onefile --name pwpInstaller")  # noqa
            print(f"msg     DONE: build exe pwpInstaller -> {res}")

        if not self.args.pypi and not self.args.gitlab:
            return

        pypi = "--Pypi" if self.args.pypi else ""
        gitlab = "--GitLab" if self.args.gitlab else ""

        res = os.system(f"python upload.py {pypi} {gitlab}")
        print(f"msg     DONE: upload .py {pypi} {gitlab} -> {res}")


def patcher_main(arguments):
    arguments = arguments or []
    print('msg     --------------- starting patcher_main')
    print("msg")
    print(f"msg     current user = '{getpass.getuser()}'")
    print(f"msg     current dir  = '{os.getcwd()}'")
    print(f"msg     HOME         = '{os.path.expanduser('~')}' ")
    print("msg")

    my_patcher = PwpPatcher(arguments=arguments)
    my_patcher.run()
    print('msg     --------------- ending patcher_main')


if __name__ == "__main__":
    sys.exit(patcher_main(sys.argv[1:]))



import setuptools
from piwiPre.pwpVersion import PwpVersion

# CAVEAT : This file is edited by pwpPatcher.py to set up the version name from the latest tag
with open("README.md", "r") as fh:
    long_description = fh.read()

# see https://setuptools.pypa.io/en/latest/references/keywords.html
setuptools.setup(
    name="piwiPre",
    version=PwpVersion.spec,
    author="Fabien BATTINI",                    # noqa
    author_email="fabien.battini@gmail.com",
    description="piwigo picture preparation tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/fabien.battini/piwipre",
    license_files=["LICENCE"],
    keywords='piwigo images pictures',
    packages=setuptools.find_packages(),
    classifiers=[
        # see https://pypi.org/classifiers/
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",  # noqa
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Office/Business :: Office Suites",
    ],
    entry_points={
        'console_scripts': [
            'piwiPre = PiwiPre.pwpConfig:pwp_main', # noqa
        ]
    },
    python_requires='>=3.6',
)


import io
import os
import pathlib
import re
from PIL import Image
from PIL import ImageTk

# TODO: use base64


class PwpBioImage:
    def __init__(self, content: io.BytesIO, img_format: str, filename: str):
        self.buffer = content
        self.img_format = img_format
        self.filename = filename

    def image(self):
        return Image.open(self.buffer)

    def tk_bitmap(self):
        return ImageTk.BitmapImage(Image.open(self.buffer))

    def tk_photo(self):
        return ImageTk.PhotoImage(Image.open(self.buffer))


class PwpEmbedder:

    def __init__(self):
        pass

    def strip(self, name: str):
        return

    @staticmethod
    def convert_data(inf, in_file: str, out_file: str, img_format: str):
        image_name = re.sub(r"[^a-zA-Z0-9_@~àâäéèêëïîôöùûüÿçñÀÂÄÉÈÊËÏÎÔÖÙÛÜŸÇÑ]",
                            "_",
                            os.path.basename(in_file))
        with open(out_file, "w", encoding="utf8") as ouf:
            print(f"# file generated by pwpImageEmbedder from {in_file}\n", file=ouf)
            print("import io\n", file=ouf, end='')
            print("from piwiPre.pwpImageEmbedder import PwpBioImage\n", file=ouf)

            line = f'{image_name} = PwpBioImage(io.BytesIO('
            header = " " * len(line)
            chunk = inf.read(2048)
            buff = b''
            while chunk:
                for i in range(0, len(chunk)):
                    if len(f"{line}{buff+chunk[i:i+1]}") > 119:
                        print(f"{line}{buff}\t# noqa",  file=ouf, end='\n')
                        line = header
                        buff = chunk[i:i+1]
                    else:
                        buff += chunk[i:i+1]
                chunk = inf.read(2048)

            print(f'{line}{buff}),\t# noqa\n', file=ouf, end='')
            line = f'{image_name} = ('
            header = " " * len(line)
            print(f'{header}img_format="{img_format}",\t# noqa\n', file=ouf, end='')
            print(f'{header}filename="{in_file}")\t# noqa\n', file=ouf, end='')

    def convert(self, in_file: str, out_file: str):
        with open(in_file, "rb") as inf:
            img_format = pathlib.Path(in_file).suffix[1:]
            self.convert_data(inf, in_file, out_file, img_format)

    def resize_and_convert(self, in_file: str, out_file: str, width: int, height: int):
        src = Image.open(in_file)
        img_format = pathlib.Path(in_file).suffix[1:]
        src.thumbnail((width, height), resample=Image.LANCZOS)
        buffer = io.BytesIO()
        src.save(buffer, format=img_format)
        b3 = io.BytesIO(buffer.getvalue())
        self.convert_data(b3, in_file, out_file, img_format)


if __name__ == "__main__":
    embedder = PwpEmbedder()
    embedder.resize_and_convert("../pwpLogo.png", "../piwiPre/pwpLogoSmall.py", 50, 29)



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini@gmail.com
# ---------------------------------------------------------------------------------------------------------------

import datetime
import logging
import re
import os
import pathlib
import subprocess

# pip install Pillow
from PIL import Image
from PIL.ExifTags import TAGS, GPSTAGS
from PIL import ImageOps

# pip install iptcinfo3
from iptcinfo3 import IPTCInfo

from piwiPre.pwpActor import ACTOR, PwpSummary
from piwiPre.pwpConfig import PwpConfig
from piwiPre.pwpErrors import LOGGER, PwpInternalError, PwpError, PwpConfigError
from piwiPre.pwpDir import PwpDirEntry, PwpFileEntry


class PwpData:
    """Root class for PwpJpg and PwpMp4"""
    # opened_objects = {}
    jpg = ("jpg", "jpeg")
    image = ("png", "gif", "bmp", "eps", "ico", "pbm", "pgm", "pnm", "tga", "tiff", "webp", "xbm", "wmf", "xpm")
    # see https://pillow.readthedocs.io/en/stable/handbook/image-file-formats.html
    mp4 = ("mp4",)
    video = ("webm", "ogv", "m4v", "mov",
             "mkv", "flv", "vob", "ogg", "avi", 'mts', 'm2ts', 'ts', "wmv", 'rm', 'rmvb', 'asf',
             "mpg", "mp2", "mpeg", "mpe", "mpv", "m2v", "m4v")
    kept = ('txt', )

    avoided = ('ini', 'db')

    # .ini file are not copied from triage to album, and kept in verify-album

    JPG = "jpg"
    IMAGE = "image"
    MP4 = "mp4"
    VIDEO = "video"
    KEPT = "kept"
    AVOIDED = "avoided"

    def __init__(self, filename, working, backup):
        """
        :param filename: the current name
        :param working: file we can modify, can be filename or a tmp file
        :param backup: where we will copy, if modifications are allowed in place
        """
        self.filename = filename
        self.working = working
        self.backup = backup
        self.width = None
        self.height = None

        self.suffix = pathlib.Path(filename).suffix[1:]
        self.representative = None          # picture that represents the file if this is a video
        self.to_abort = False

        self.can_be_inserted_in_db = False  # will be set to True for images and video

        if filename and ACTOR.isfile(filename):
            self.size = os.stat(filename).st_size
            self.modified_time = os.path.getmtime(filename)
        else:
            self.size = None
            self.modified_time = datetime.datetime.now().timestamp()

        # metadata from file JPG, MP4, etc metadata inside the file
        self.copyright = None
        self.special = None
        self.author = None
        self.creation = None
        self.model = None
        self.latitude = None
        self.longitude = None

        # these metadata exists only for JPG
        self.iptc_info = None
        self.exif_data = None

        self.is_modified = False

    def __enter__(self):
        return self  # everything done in __init__

    def __exit__(self, _exc_type, _exc_val, _exc_tb):
        self.close()

    def incr_logger_count(self, item=None):
        raise PwpInternalError("incr_logger_count()")

    @staticmethod
    def get_type(filename):
        suffix = pathlib.Path(filename).suffix[1:].lower()

        if suffix in PwpData.jpg:
            return PwpData.JPG

        if suffix in PwpData.image:
            return PwpData.IMAGE

        if suffix in PwpData.mp4:
            return PwpData.MP4

        if suffix in PwpData.video:
            return PwpData.VIDEO

        if suffix in PwpData.kept:
            return PwpData.KEPT

        return PwpData.AVOIDED

    @staticmethod
    def create(filename, config: PwpConfig,  stage,
               tmp: str or None = None,
               backup: str or None = None):
        """
        Decides the flavor of PwpFile to build, depending on filename suffix.

        :param filename: the current name of the file
        :param tmp: where we will copy, if modifications are not allowed in place
        :param backup: where we will copy, if modifications are allowed in place
        :param stage: processing stage: triage or album
        :param config: current configuration
        :returns: a PwpFile, opened.
        """

        file_type = PwpData.get_type(filename)

        if file_type == PwpData.JPG:
            backup = None if stage == "triage" else backup
            working = tmp
            return PwpJpg(filename, config=config, working=working, backup=backup)

        if file_type == PwpData.IMAGE:
            # if we have to modify, then we need to back up
            if tmp:
                # this happens when we manage a png with enable-rename false
                le = len(pathlib.Path(tmp).suffix[1:])
                tmp = tmp[:-le] + 'jpg'
                # here, we FORCE working to be a JPG file
            return PwpImage(filename, config=config, working=tmp, backup=backup)

        if file_type == PwpData.MP4:
            backup = None if stage == "triage" else backup
            working = tmp
            return PwpMp4(filename, config=config, working=working, backup=backup)

        if file_type == PwpData.VIDEO:
            # if we have to modify, then we need to back up
            if tmp:
                le = len(pathlib.Path(tmp).suffix[1:])
                tmp = tmp[:-le] + 'mp4'
            return PwpVideo(filename, config=config, working=tmp, backup=backup)

        if file_type == PwpData.KEPT:
            return PwpKept(filename)

        return PwpAvoided(filename)

    def close(self):
        # self.release(force_close=False)
        pass  # nothing to do on txt files

    def close_if_backup(self):
        """
        closes the file BEFORE it is modified.
        if backup is set, copies the file to back up
        :return: None
        """
        if self.backup:
            LOGGER.info(f"backup {self.filename} to {self.backup}")
            ACTOR.copy(self.filename, self.backup)
            self.close()

    def patch_after_rename(self, new_filepath):
        """
        For some objects, it is necessary to patch the name that rename generates
        :param new_filepath:
        :return: patched version
        """
        raise PwpInternalError(f"patch_after_rename({self.filename})")
        # return new_filepath

    def save_with_info(self, _config, new_iptc=None):
        raise PwpInternalError(f"save_with_info({self.filename})")
        # pass  # nothing to do on txt files

    def handles_metadata(self):
        raise PwpInternalError(f"handles_metadata({self.filename})")
        #  return False

    def write_metadata(self, _new_author, _new_copyright, _new_special, _new_date, _config):
        raise PwpInternalError(f"write_metadata({self.filename})")
        #  return self

    def rotate(self, _stage: str, _any):
        raise PwpInternalError(f"rotate({self.filename})")
        # return self  # nothing to do with text or mp4

    @staticmethod
    def str_ptime(str_rep, str_format):    # noqa
        try:
            return datetime.datetime.strptime(str_rep, str_format)
        except ValueError:
            return None

    @staticmethod
    def datetime(filename, year, month, day, hour, minute, second):
        try:
            from_filename = datetime.datetime(year=year,
                                              month=month,
                                              day=day,
                                              hour=hour,
                                              minute=minute,
                                              second=second)
            return from_filename
        except ValueError as e:
            LOGGER.info(f"Error {e} in date from filename {filename} ")
            LOGGER.info(f"year={year}  month={month} day={day}")
            LOGGER.info(f"hour={hour}  minute={minute} second={second}")
            return None

    @staticmethod
    def get_date(str_rep):
        if str_rep is None:
            raise PwpInternalError("get_date(None)")

        r = re.match(r"\d\d\d\d:\d\d:\d\d \d\d:\d\d:\d\d", str_rep)
        if r and str_rep != '0000:00:00 00:00:00':
            return PwpData.str_ptime(str_rep, '%Y:%m:%d %H:%M:%S')

        r = re.match(r"\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d", str_rep)
        if r:
            return PwpData.str_ptime(str_rep, '%Y-%m-%d %H:%M:%S')

        # '2023-01-27T17:59:39.000000Z' strptimes has difficulties, there  # noqa
        r = re.match(r"(\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d)", str_rep)
        if r:
            return datetime.datetime(year=int(r.group(1)),
                                     month=int(r.group(2)),
                                     day=int(r.group(3)),
                                     hour=int(r.group(4)),
                                     minute=int(r.group(5)),
                                     second=int(r.group(6)))

        # r = re.match(r"\d\d\d\d/\d\d/\d\d \d\d:\d\d:\d\d", str_rep)
        # if r:
        #     return PwpFile.str_ptime(str_rep, '%Y/%m/%d %H:%M:%S')
        raise PwpInternalError(f"get_date({str_rep})")

    @staticmethod
    def date_from_dico(filename, file_info) -> datetime or None:
        """
        date_from_dico(file_info)
        :param filename: name of the file
        :param file_info: dictionary, an output from ACTOR.get_info_from_format
        :return: datetime.datetime
        """
        if file_info is None or\
                'Y' not in file_info or file_info['Y'] is None or \
                'm' not in file_info or file_info['m'] is None or \
                'd' not in file_info or file_info['d'] is None:
            return None

        if file_info['H']:
            hour = int(file_info['H'])
        elif 'count' in file_info and file_info['count']:
            hour = int(int(file_info['count']) / 3600)
        else:
            hour = 0

        if file_info['M']:
            minute = int(file_info['M'])
        elif 'count' in file_info and file_info['count']:
            minute = int(int(file_info['count']) / 60)
        else:
            minute = 0

        if file_info['S']:
            second = int(file_info['S'])
        elif 'count' in file_info and file_info['count']:
            second = int(file_info['count']) % 60
        else:
            second = 0

        return PwpData.datetime(filename,
                                year=int(file_info['Y']),
                                month=int(file_info['m']),
                                day=int(file_info['d']),
                                hour=hour,
                                minute=minute,
                                second=second)

    def guess_file_date(self, image_path: str, config: PwpConfig, default_to_filedate=True)\
            -> datetime or None:
        """
        :param image_path: the full path to image
        :param config: the current configuration
        :param default_to_filedate: if True and no date found, returns the file date
        :return: date
        """  # noqa

        filename = os.path.basename(image_path)
        template = os.path.basename(config['names'])
        # check if the date can be extracted from the filename, according to current template
        file_info = ACTOR.get_info_from_format(template, filename)
        from_filename = self.date_from_dico(filename, file_info)

        if from_filename:
            return from_filename

        # TODO: Manage timezone and AM/PM

        m = re.match(r"VID(\d\d\d\d)(\d\d)(\d\d)(\d\d)(\d\d)(\d\d)\.mp4", filename) or \
            re.match(r"VID_(\d\d\d\d)(\d\d)(\d\d)_(\d\d)(\d\d)(\d\d)\.mp4", filename) or \
            re.match(r".*-(\d\d\d\d)-(\d\d)-(\d\d)-(\d\d)h(\d\d)-(\d\d).*\.jpg", filename)
        if m:  # pragma: no cover
            year = int(m.group(1))
            month = int(m.group(2))
            day = int(m.group(3))
            hours = int(m.group(4))
            minutes = int(m.group(5))
            seconds = int(m.group(6))
            from_filename = PwpJpg.datetime(filename, year=year, month=month, day=day,
                                            hour=hours, minute=minutes, second=seconds)
            if from_filename:
                return from_filename
            # in real life, this can happen, but we have (yet) no test case for this

        m = re.match(r"IMG-(\d\d\d\d)-(\d\d)(\d\d)-W.(\d\d)(\d\d).*\.jpg", filename) or \
            re.match(r"IMG-(\d\d\d\d)(\d\d)(\d\d)-W.(\d\d)(\d\d).*\.jpg", filename)
        # for WhatsApp images, which does not have IPTC date information, but get a date in the picture name
        if m:
            year = int(m.group(1))
            month = int(m.group(2))
            day = int(m.group(3))
            hours = int(m.group(4)) + 12
            minutes = int(m.group(5))
            seconds = 0
            from_filename = PwpData.datetime(filename, year=year, month=month, day=day,
                                             hour=hours, minute=minutes, second=seconds)
            if from_filename:
                return from_filename

        # maybe the date can be extracted from the directory, and time from the file
        if default_to_filedate and ACTOR.isfile(image_path):
            template = os.path.basename(os.path.dirname(config['names']))  # the 1st item in dir hierarchy
            dir_name = os.path.basename(os.path.dirname(image_path))
            dir_info = ACTOR.get_info_from_format(template, dir_name)
            if dir_info is not None:
                file_date = datetime.datetime.fromtimestamp(os.path.getmtime(image_path))

                dir_info['H'] = file_date.hour
                dir_info['M'] = file_date.minute
                dir_info['S'] = file_date.second
                from_filename = self.date_from_dico(filename, dir_info)

                if from_filename:
                    return from_filename

        if default_to_filedate and ACTOR.isfile(image_path):
            LOGGER.debug(f"Cannot guess date for {image_path}")
            return datetime.datetime.fromtimestamp(os.path.getmtime(image_path))
        return None

    @staticmethod
    def date_to_str(c_date):
        return f"{c_date.year:04}/{c_date.month:02}/{c_date.day:02} " + \
            f"{c_date.hour:02}:{c_date.minute:02}:{int(c_date.second):02}"

    def compute_new_date_author(self, config: PwpConfig):
        """
        :param config: current configuration
        :return: date, author
        """

        from_filename = self.guess_file_date(self.filename, config, default_to_filedate=False)
        if from_filename and config['trust-date-in-filename']:
            author = config.author(self.model, from_filename)
            return from_filename, author

        from_exif = self.creation
        if from_exif:
            LOGGER.debug(f"Date = {self.date_to_str(from_exif)}  before correction")
        else:
            LOGGER.debug("Date = None before correction")

        corr_fr_exif = config.fix_date(self.filename, from_exif, self.model)

        if corr_fr_exif:
            LOGGER.debug(f"Date = {self.date_to_str(corr_fr_exif)} after correction")

            author = config.author(self.model, corr_fr_exif)
            return corr_fr_exif, author
        else:
            LOGGER.debug("Date = None after correction")

        if from_filename:
            author = config.author(self.model, from_filename)
            return from_filename, author

        LOGGER.debug(f"No exif nor file dates in '{self.filename}'")

        file_date = datetime.datetime.fromtimestamp(os.path.getmtime(self.filename))
        author = config.author(self.model, file_date)
        return file_date, author

    def verify_metadata(self, _stage,
                        config: PwpConfig,
                        summary: PwpSummary,
                        new_date: datetime,
                        new_author="Photographer",
                        enable_reset=True):
        """
        :param _stage: triage or album
        :param config: local configuration
        :param new_date: date to set
        :param new_author: author to set in exif
        :param summary: remembers the actions taken, for debug
        :param enable_reset: is True, always write metadata, if False, write only if not set

                self is supposed to be open

        if necessary:
        - saves the file to back up
        - rotates the file
        - stores the rotated picture in a copy
        - set is_modified to True
        - manages LOGGER.info()

        lives the file open for further processing

        :return self or a copy, modified"""

        if config['enable-metadata'] is False or self.handles_metadata() is False:
            return self, False

        is_modified = False
        if (enable_reset and self.creation != new_date) or self.creation is None or self.creation == '':
            summary.meta_date = True
            is_modified = True

        dico = config.format_dict(new_date, new_author)

        info_format = config.format('instructions')

        new_special = info_format.format(**dico)
        if (enable_reset and self.special != new_special) or self.special is None or self.special == '':
            summary.meta_instructions = True
            is_modified = True

        if (enable_reset and self.author != new_author) or self.author is None or self.author == '':
            summary.meta_author = True
            is_modified = True

        copyright_format = config.format('copyright')
        new_copyright = copyright_format.format(**dico)
        if (enable_reset and self.copyright != new_copyright) or self.copyright is None or self.copyright == '':
            summary.meta_copyright = True
            is_modified = True

        if not is_modified:
            return self, is_modified

        modifications = summary.get_meta()

        # LOGGER.msg(f"write IPTC info '{self.filename}'")
        if config['dryrun']:
            LOGGER.info(f"Would insert meta[{modifications}] metadata in '{self.working}'")
            return self, is_modified

        if self.is_modified:
            # this happens if the file was rotated
            # saving it is useless, and we already have a copy
            pass
        else:
            # here, the file WILL be modified
            self.close_if_backup()

        new_image = self.write_metadata(new_author, new_copyright, new_special, new_date, config)
        new_image.incr_logger_count("Metadata")

        LOGGER.info(f"metadata meta[{modifications}] inserted in '{new_image.filename}'")

        return new_image, is_modified

    @staticmethod
    def get_remote_web_filename(config, filename):
        """
        :param config: local configuration
        :param filename: local name
        :return: the remote path to be used for sftp
        """
        if config['enable-remote-web'] and ACTOR.ssh_connection:
            remote_file = filename.replace(config['web'], config['remote-web'])
            return remote_file
        return None

    def thumbnail(self, filename, width, height, crop, config: PwpConfig, force_rebuild: bool,
                  thumbs_dir: PwpDirEntry, summary: PwpSummary):
        # no thumbnail generated
        return

    def verify_orientation(self, _stage: str, _config: PwpConfig, _summary: PwpSummary):
        return self, None

# ------------------------------------------------------------------------------------------------
# PwpImage
# ------------------------------------------------------------------------------------------------


class PwpImage(PwpData):
    """Class for images that will be moved to Jpg"""

    def __init__(self, filename, config: PwpConfig,
                 working: str or None = None,
                 backup: str or None = None,
                 image=None, iptc_info=None):
        super().__init__(filename, working=working, backup=backup)
        self.creation = self.guess_file_date(filename, config, default_to_filedate=True)

        self.can_be_inserted_in_db = True

        self.iptc_info = iptc_info
        self.image = image
        if image is not None:
            self.image = image
        elif self.filename is not None:
            self.image = Image.open(self.filename)
            if self.image is None:
                raise PwpError(f"error opening image {self.filename}")

        self.width, self.height = self.image.size

    def incr_logger_count(self, item=None):
        LOGGER.incr_picture("Image" + (f": {item}" if item else ""))

    def close(self):
        """close the image WITHOUT saving"""
        if self.image:
            self.image.close()
            super().close()
            # is self.image is None, the close has already been done
        self.image = None
        if self.filename and ACTOR.isfile(self.filename + '~'):
            ACTOR.delete(self.filename + '~')

    def reopen(self):
        if self.image is None:
            if not ACTOR.isfile(self.filename):
                raise PwpInternalError(f"reopen('{self.filename}'")
            self.image = Image.open(self.filename)

    # We never need patch_after_rename for images,
    # because ALL images are silently converted to JPG in verify_orientation or verify_metadata
    # because we do not read metadata of non JPG files, we do not find metadata,
    # and verify_metadata always generate a JPG file

    def save_with_info(self, config: PwpConfig, new_iptc=None):
        if config['dryrun']:
            LOGGER.debug(f"would save image '{self.filename}'")
        else:
            # need to create the file BEFORE setting the iptc info
            if self.image is None:
                raise PwpInternalError("Image is None")
            ACTOR.mkdirs(os.path.dirname(self.filename))
            self.image.save(self.filename, exif=self.exif_data)
            self.image.close()
            if new_iptc:
                info_resized = IPTCInfo(self.filename, force=True)
                info_resized['copyright notice'] = new_iptc['copyright notice']
                info_resized['special instructions'] = new_iptc['special instructions']
                info_resized['by-line'] = new_iptc['by-line']
                self.iptc_info = info_resized

            if self.iptc_info:
                self.iptc_info.save()

        self.is_modified = False
        self.close()

    def verify_orientation(self, stage: str, config: PwpConfig, summary: PwpSummary):
        ACTOR.copy(self.filename, self.working, forced=True)
        # here, we copy self to working, which DOES have a .jpg extension
        # so, we MAY end-up with a PNG data within a JPG filename
        # this is NOT an issue, because working WILL be modified and saved,
        # and then it WILL have the right JPG format !
        # not very clean, but works.
        # but DOES NOT TAKE enable-conversion into account.

        if stage == "album":
            if config['dryrun']:
                ACTOR.copy(self.filename, self.backup, forced=True)
            else:
                ACTOR.move(self.filename, self.backup, forced=True)
            LOGGER.info(f"backup {self.filename} to {self.backup}")
        if self.working[-4:] == ".jpg":
            new_image = PwpJpg(filename=self.working,
                               working=self.working,
                               backup=None,
                               config=config)  # because the original file has already been back upped
            return new_image.verify_orientation(stage, config, summary)
        else:
            # if this is not a JPG, we DO NOT verify the orientation

            # CAVEAT: This part of the code (i.e. when enable-conversion = false) is NOT Tested
            #
            new_image = PwpData.create(filename=self.working,
                                       config=config,
                                       stage=stage,
                                       tmp=self.working,
                                       backup=None)  # because the original file has already been back upped
            return new_image

    def thumbnail(self, filename, width, height, crop, config: PwpConfig, force_rebuild: bool,
                  thumbs_dir: PwpDirEntry, summary: PwpSummary):
        """
        Build a thumbnail and copy it to remote if necessary.
        The web files (thumbnails, .ini from auto-config) are created from local pictures.

        - If a local thumbnail exist and is more recent than the picture,
          then it is considered as valid, and put to the remote location
        - in all other cases, the thumbnail is created again, and put to the remote location afterward

        :param filename: filename name of the thumbnail
        :param width: expected max width
        :param height: expected max height
        :param crop: boolean
        :param config: current configuration
        :param force_rebuild: True, False. If True, MUST rebuild the thumbnail.
        :param thumbs_dir: DirEntry of the directory
        :param summary:
        :return: None
        """

        def after_creation(created: bool):
            """
            Executes actions after the creation of the thumbnail:
            - clean filename~
            - put to remote if necessary
            - setup summary
            :return: None
            """
            pwp_file = PwpFileEntry.lookup(filename, context="web", config=config)

            if ACTOR.isfile(self.filename + '~'):
                ACTOR.delete(self.filename + '~')

            suf = filename[-6:-4]
            if created:
                if suf == "50":
                    summary.thumb_cu = True
                elif suf == "th":
                    summary.thumb_th = True
                elif suf == "me":
                    summary.thumb_me = True
                elif suf == "sq":
                    summary.thumb_sq = True
                else:
                    raise PwpInternalError(f"Illegal thumbnail {filename}")

            if pwp_file.put():
                remote_path = os.path.dirname(pwp_file.remote)
                LOGGER.info(f"Thumbnail {filename} sftp to {remote_path}")
                if suf == "50":
                    summary.rem_thumb_cu = True
                elif suf == "th":
                    summary.rem_thumb_th = True
                elif suf == "me":
                    summary.rem_thumb_me = True
                elif suf == "sq":
                    summary.rem_thumb_sq = True
                else:
                    raise PwpInternalError(f"Illegal thumbnail {filename}")

            self.incr_logger_count("Thumbnail")
            # --------------------------------------- end of subroutine

        if filename in thumbs_dir.files and force_rebuild:
            del thumbs_dir.files[filename]

        if thumbs_dir.exists_and_younger_than(filename, datetime.datetime.fromtimestamp(self.modified_time)):
            LOGGER.info(f"Thumbnail {filename} is more recent than {self.filename}")
            after_creation(created=False)
            return

        if config['dryrun']:
            LOGGER.info(f"Would create Thumbnail {width}x{height} crop={crop} name={filename}")
            return

        LOGGER.info(f"Thumbnail {width}x{height} crop={crop} {filename}")

        self.reopen()
        if not crop:
            im_resized = PwpJpg(filename, working=None, backup=None,
                                image=self.image.copy(), iptc_info=self.iptc_info,
                                config=config)
            im_resized.image.thumbnail((width, height), resample=Image.LANCZOS)
        else:
            im_resized = PwpJpg(filename, working=None, backup=None,
                                image=ImageOps.fit(self.image, (width, height)),
                                iptc_info=self.iptc_info,
                                config=config)

        dir_name = os.path.dirname(filename)

        ACTOR.mkdirs(dir_name)
        im_resized.set_exif_data('ImageWidth', im_resized.image.width)
        im_resized.set_exif_data('ImageLength', im_resized.image.height)

        if not self.iptc_info:
            # if the source image is not JPG and enable-rename false,
            # then self is not a JPG and self.iptc_info is empty
            raise PwpConfigError(f"enable-rename is false, so '{filename}' cannot be safely managed")

        new_iptc = {'copyright notice': self.iptc_info['copyright notice'] or [],
                    'special instructions': self.iptc_info['special instructions'] or [],
                    'by-line': self.iptc_info['by-line'] or []
                    }

        im_resized.is_modified = True
        im_resized.save_with_info(config, new_iptc=new_iptc)

        after_creation(created=True)

    def handles_metadata(self):
        return True


# ------------------------------------------------------------------------------------------------
# PwpJpg
# ------------------------------------------------------------------------------------------------


class PwpJpg(PwpImage):
    logger = logging.getLogger('iptcinfo')
    logger.setLevel(logging.ERROR)
    tags_reverse = dict(((v, k) for k, v in TAGS.items()))
    orientations = ((None, False),  # 0: is an error
                    (None, False),  # 1: nothing to do
                    (Image.FLIP_LEFT_RIGHT, False),  # 2
                    (Image.ROTATE_180, False),  # 3
                    (Image.FLIP_TOP_BOTTOM, False),  # 4
                    (Image.ROTATE_270, True),  # 5
                    (Image.ROTATE_270, False),  # 6
                    (Image.ROTATE_90, True),  # 7
                    (Image.ROTATE_90, False))  # 8
    orient_names = (('None', 'No flip'),  # 0: is an error
                    ('None', 'No flip'),  # 1: nothing to do
                    ('FLIP_LEFT_RIGHT', 'No flip'),  # 2
                    ('ROTATE_180', 'No flip'),  # 3
                    ('FLIP_TOP_BOTTOM', 'No flip'),  # 4
                    ('ROTATE_270', 'Flip'),  # 5
                    ('ROTATE_270', 'No flip'),  # 6
                    ('ROTATE_90', 'Flip'),  # 7
                    ('ROTATE_90', 'No flip'))  # 8
    orient_summa = ('--',  # 0: is an error
                    '--',  # 1: nothing to do
                    '|-',  # 2
                    '^-',  # 3
                    'V-',  # 4
                    '<|',  # 5
                    '<-',  # 6
                    '>|',  # 7
                    '>-')  # 8

    def __init__(self, filename, config: PwpConfig, working=None, backup=None, image=None, iptc_info=None):
        super().__init__(filename, config, working, backup,  image=image, iptc_info=iptc_info)
        self.exif_data = None

        self.sub_sec_time = None
        self.orientation = 1

        self.latitude = None
        self.longitude = None

        # NB: width, height = self.image.size

        # the values initialized here are the INITIAL values BEFORE piwiPre actions
        self.exif_data = self.image.getexif()
        if self.exif_data:
            exif_table = {}
            gps_data = {}
            for tag, value in self.exif_data.items():
                decoded = TAGS.get(tag, tag)
                exif_table[decoded] = value
                if decoded == 'GPSInfo':
                    gps_info = self.exif_data.get_ifd(tag)
                    for key, val in gps_info.items():
                        sub_decoded = GPSTAGS.get(key, key)
                        gps_data[sub_decoded] = val

            date_time = exif_table['DateTime'] if 'DateTime' in exif_table else \
                exif_table['DateTimeOriginal'] if 'DateTimeOriginal' in exif_table else None

            self.creation = self.get_date(date_time)

            self.make = exif_table['Make'] if 'Make' in exif_table else None
            self.model = exif_table['Model'] if 'Model' in exif_table else None
            self.sub_sec_time = exif_table['SubsecTime'] if 'SubsecTime' in exif_table else None   # noqa
            self.orientation = int(exif_table['Orientation']) if 'Orientation' in exif_table else 1

            self.latitude = self.get_gps(gps_data, 'GPSLatitudeRef', 'GPSLatitude')
            self.longitude = self.get_gps(gps_data, 'GPSLongitudeRef', 'GPSLongitude')

            # sst2 = self.get_data(exif_data, 'SubsecTimeOriginal')     # noqa
            # sst3 = self.get_data(exif_data, 'SubsecTimeDigitized')    # noqa
        else:
            LOGGER.debug(f"No exif data in '{self.filename}'")

        if iptc_info:
            self.iptc_info = iptc_info
        elif self.filename and ACTOR.isfile(self.filename):
            self.iptc_info = IPTCInfo(self.filename, force=True)

        if isinstance(self.iptc_info, IPTCInfo):
            b1 = self.iptc_info['special instructions']
            self.special = self.decode(b1, self.filename, "special")
            b2 = self.iptc_info['copyright notice']
            self.copyright = self.decode(b2, self.filename, "copyright")
            b3 = self.iptc_info['by-line']
            self.author = self.decode(b3, self.filename, "author")

        #  self.close() # NO: we keep the file open because we may want to rotate it.

    @staticmethod
    def decode(buffer, filename, info):
        if not buffer:
            return None
        try:
            return buffer.decode('utf8')
        except UnicodeDecodeError as e:
            LOGGER.info(f"Error {e} while decoding IPTC {info} for {filename}")
        return None

    def incr_logger_count(self, item=None):
        LOGGER.incr_picture("Jpg" + (f": {item}" if item else ""))

    def patch_after_rename(self, new_filepath):
        """
        For some objects, it is necessary to patch the name that rename generates
        :param new_filepath:
        :return: patched version
        """
        return new_filepath

    def handles_metadata(self):
        return True

    # management of GPS data in EXIF:
    # exif standard can be found on http://web.archive.org/web/20131018091152/http://exif.org/Exif2-2.PDF
    # Name              TagID (hex)     Type        Count
    # GPSLatitudeRef    1               ascii       2       'N' for north, 'S' for south
    # GPSLatitude       2               RATIONAL    3
    # GPSLongitudeRef   3               ascii       2       'E'/'W'
    # Longitude         4               RATIONAL    3
    #
    # Rational:
    #   method 1:   degrees, minutes and second : dd/1,mm/1,ss/1
    #   method 2:   degrees and minutes :  dd/1,mmmm/100,0/1.

    @staticmethod
    def get_gps(data: dict, ref, rational):
        """
        decodes from DMS or DM to float
        :param data: dict of gps data
        :param ref: string for N/S or E/W reference
        :param rational: string for 3 valued data
        :return: lat (or long)
        """
        if rational not in data.keys() or ref not in data.keys():
            return None

        sign = -1 if data[ref] in ['S', 'W'] else 1
        deg, minutes, sec = data[rational]
        return sign * round(float(deg + minutes/60 + sec / 3600), 6)  # 6 is te precision of floats in the database

    def set_exif_data(self, tag_id, value):
        # https://www.thepythoncode.com/code/extracting-image-metadata-in-python
        # get the tag name, instead of human unreadable tag id
        tag = self.tags_reverse.get(tag_id)
        self.exif_data[tag] = value

    def write_metadata(self, new_author: str, new_copyright: str, new_special: str,
                       new_date: datetime.datetime,
                       config: PwpConfig):
        if not self.iptc_info:
            raise PwpError(f"ERROR writing IPTC from '{self.filename}'")

        if self.working != self.filename:
            ACTOR.copy(self.filename, self.working)

            new_image = PwpJpg(filename=self.working,
                               working=self.working,
                               backup=None,
                               config=config)            # because the original file has already been back upped
        else:
            new_image = self
            self.reopen()

        new_author = new_author or ''
        new_special = new_special or ''
        new_copyright = new_copyright or ''

        new_image.special = new_special
        new_image.copyright = new_copyright
        new_image.author = new_author
        new_image.creation = new_date

        new_iptc = {'copyright notice': new_copyright.encode("utf8"),
                    'special instructions': new_special.encode("utf8"),
                    'by-line': new_author.encode("utf8")
                    }
        date_str = new_date.strftime("%Y-%m-%d %H:%M:%S")
        new_image.set_exif_data('DateTime', date_str)
        new_image.set_exif_data('DateTimeOriginal', date_str)

        new_image.is_modified = True   # to ensure file IS writen
        new_image.save_with_info(config, new_iptc)
        return new_image

    def verify_orientation(self, _stage: str, config: PwpConfig, summary: PwpSummary):
        """
        self is supposed to be open

        if necessary:
        - saves the file to back up
        - rotates the file
        - stores the rotated picture in a copy
        - set is_modified to True
        - manages LOGGER.info()

        lives the file open for further processing
        return: self or a copy, modified"""

        if config['enable-rotation'] is False:
            return self, "", False

        (angle, flip) = self.orientations[self.orientation]
        (a, r) = self.orient_names[self.orientation]
        summary.rotation = self.orient_summa[self.orientation]

        modified = False
        new_image = self
        # here, new_image is always the image we want to handle

        if angle is not None:
            if config['dryrun']:
                LOGGER.info(f"Would clean ROTATION {self.orientation} {a} {r} '{self.filename}'")
            else:
                transposed = self.image.transpose(method=angle)
                width = self.image.width
                height = self.image.height
                self.close_if_backup()

                new_image = PwpJpg(filename=self.working,
                                   working=self.working,
                                   backup=None,  # because the original file has already been back upped
                                   image=transposed, iptc_info=self.iptc_info,
                                   config=config)
                new_image.set_exif_data('ImageWidth', width)
                new_image.set_exif_data('ImageLength', height)
                new_image.set_exif_data('Orientation', 1)
                new_image.is_modified = True

                LOGGER.info(f"ROTATION {self.orientation} {a} {r} '{self.filename}'")
                modified = True

        if flip:
            if config['dryrun']:
                LOGGER.info(f"Would clean FLIP {self.orientation} {a} {r}'{self.filename}'")
            else:
                transposed = new_image.image.transpose(method=Image.FLIP_LEFT_RIGHT)
                new_image.close_if_backup()

                new_image = PwpJpg(filename=self.working,
                                   working=self.working,
                                   backup=None,  # because the original file has already been back upped
                                   image=transposed,
                                   iptc_info=self.iptc_info,
                                   config=config)
                new_image.set_exif_data('Orientation', 1)
                new_image.is_modified = True
                LOGGER.info(f"FLIP {self.orientation} {a} {r}'{self.filename}'")

                modified = True

        return new_image, modified


# ------------------------------------------------------------------------------------------------
# PwpVideo
# ------------------------------------------------------------------------------------------------


class PwpVideo(PwpData):
    """
    PwpVideo: the base class for all video files, supported or not
    uses ffmpeg and ffproble to set/get metadata
    """  # noqa

    def __init__(self, filename, working, backup, config: PwpConfig):
        super().__init__(filename, working, backup)
        self.get_metadata(config)
        self.can_be_inserted_in_db = True
        self.representative = os.path.dirname(filename) + '/pwg_representative/' + pathlib.Path(filename).stem + '.jpg'

    def incr_logger_count(self, item=None):
        LOGGER.incr_picture("Video" + (f": {item}" if item else ""))

    def handles_metadata(self):
        return True

    def get_metadata(self, config: PwpConfig):
        ffprobe = config['ffmpeg-path'] + '/ffprobe'
        cmd = [ffprobe, "-hide_banner", "-i", self.filename]
        try:
            res = subprocess.run(cmd, capture_output=True, check=True)
        except subprocess.CalledProcessError as e:
            raise PwpInternalError(f"ERROR {e} while getting metadata for '{self.filename}'")

        buffer = (res.stdout.decode('utf8') + res.stderr.decode('utf8')).splitlines()
        for line in buffer:
            r = re.match(r"\s\s\s\sartist\s*: (.*)$", line)
            if r:
                self.author = r.group(1)
                continue

            r = re.match(r"\s\s\s\scomment\s*: (.*)$", line)
            if r:
                self.special = r.group(1)
                continue

            r = re.match(r"\s\s\s\scopyright\s*: (.*)$", line)
            if r:
                self.copyright = r.group(1)
                continue

            r = re.match(r"\s\s\s\screation_time\s*: (.*)$", line)
            if r:
                str_rep = r.group(1)
                self.creation = self.get_date(str_rep)
                continue

            r = re.match(r"\s+Stream .* Video: .*, (\d+)x(\d+).*, \d+ kb/s, .*$", line)
            #  Stream #0:0[0x1](und): Video: h264 (High 4:2:2) (avc1 / 0x31637661), ..., 320x240, 546 kb/s, ...
            #  Stream #0:0[0x1](eng): Video: h264 (High)..., 1080x1920 [SAR 1:1 DAR 9:16], 5819 kb/s, ...
            if r:
                self.width = int(r.group(1))
                self.height = int(r.group(2))
                continue
        LOGGER.debug(f"Video '{self.filename}' width {self.width} height {self.height}")

    def write_metadata(self, new_author: str, new_copyright: str, new_special: str,
                       new_date: datetime.datetime, config: PwpConfig):
        self.close_if_backup()
        ACTOR.mkdirs(os.path.dirname(self.working), forced=True)

        modified = False
        path = config['ffmpeg-path'] or ""
        if path and path[-1] != "/":
            path = path + '/'
        ffmpeg = path + 'ffmpeg'
        cmd = [ffmpeg, "-i", self.filename,
               "-loglevel", "error",
               "-err_detect", "ignore_err",
               "-movflags", "+faststart",   # see https://ffmpeg.org/ffmpeg-formats.html fragmentation  # noqa
               "-c:v", "libx264",
               "-profile:v", "high",
               "-c:a", "aac",
               #  "-b:a", "128k", #  set audio bitrate to mp4 default                                   # noqa
               #  "-ar", "48k",   # set the audio sampling frequency
               "-pix_fmt", "yuvj420p",  # videoJS does not support 4:2:2 with firefox, but does with chrome... # noqa
               # this raises a warning,
               # color range should be specified with a separate flag, but doc still unclear
               ]
        if new_author:
            cmd.extend(['-metadata', f'artist={new_author}'])
            modified = True
        if new_special:
            cmd.extend(['-metadata', f'comment={new_special}'])
            modified = True
        if new_copyright:
            cmd.extend(['-metadata', f'copyright={new_copyright}'])
            modified = True
        if new_date:
            str_date = new_date.strftime("%Y-%m-%d %H:%M:%S.000000Z")  # To check: "%f" instead of 0000
            cmd.extend(['-metadata', f'creation_time={str_date}'])
            modified = True

        # example cmd:
        # "C:/Program Files (x86)/ffmpeg/bin/ffmpeg.exe" -i "2016-06-18_Gym_144055.mp4" -loglevel error           # noqa
        #  -movflags +faststart -c:v libx264  -profile:v main -c:a aac -ar 48k -b:a 128k -pix_fmt yuvj420p # noqa
        #  -metadata author="Author"  "\\NAS\photo\Gym\2016-06-18-Gym\2016-06-18_Gym_144055.mp4"                  # noqa

        cmd.append(self.working)

        if not modified:
            return self

        # if the original file is mp4 and has not been already processed by piwiPre
        # then there is almost no chance that it already has all metadata
        # so we end-up converting it again, which adds the faststart mp4 option
        try:
            subprocess.run(cmd, capture_output=False, check=True)
        except subprocess.CalledProcessError as e:
            raise PwpInternalError(f"ERROR {e} while transcoding '{self.filename}' -> '{self.working}'")

        return PwpData.create(self.working,
                              config=config,
                              stage="triage",       # NB: triage or album do not make a difference, here
                              tmp=self.working,
                              backup=None)

    def verify_orientation(self, _stage: str, _config: PwpConfig, _summary: PwpSummary):
        return self, False

    def patch_after_rename(self, new_filepath):
        """
        For KEPT objects, it is necessary to patch the name that rename generates:
        :param new_filepath: after rename
        :return: patched version
        """
        old_filename = os.path.basename(self.filename)
        suffix = pathlib.Path(old_filename).suffix[1:]
        le = len(suffix)
        new_filepath = new_filepath[:-le] + 'mp4'
        return new_filepath

    @staticmethod
    def build_representative(source, destination, config: PwpConfig,
                             new_author: str, new_copyright: str, new_special: str,
                             new_date: datetime.datetime):
        #
        # ffmpeg  -i input -loglevel error- vf select='eq(pict_type\,I)'  -frames:v 1 -q:v 2 output.jpg
        #
        path = config['ffmpeg-path'] or ""
        if path and path[-1] != "/":
            path = path + '/'
        ffmpeg = path + 'ffmpeg'
        cmd = [ffmpeg,
               "-i", source,
               "-loglevel", "error",
               "-vf", "select='eq(pict_type\\,I)'",
               "-frames:v", "1",
               "-q:v", "2",
               destination,
               ]
        if new_author:
            cmd.extend(['-metadata', f'artist={new_author}'])
        if new_special:
            cmd.extend(['-metadata', f'comment={new_special}'])
        if new_copyright:
            cmd.extend(['-metadata', f'copyright={new_copyright}'])
        if new_date:
            str_date = new_date.strftime("%Y-%m-%d %H:%M:%S.000000Z")  # To check: "%f" instead of 0000
            cmd.extend(['-metadata', f'creation_time={str_date}'])

        try:
            subprocess.run(cmd, capture_output=True, check=True)
        except subprocess.CalledProcessError as e:
            raise PwpInternalError(f"ERROR {e} while building representative for '{source}'")

        jpg = PwpData.create(destination, stage="album", config=config, tmp=destination)
        # working = destination insures that the metadata is writen correctly

        # stage = album because representative is always inside album
        jpg.write_metadata(new_author=new_author, new_copyright=new_copyright, new_special=new_special,
                           new_date=new_date, config=config)
        jpg.close()
        LOGGER.info(f"Built representative '{destination}' for '{source}'")


class PwpMp4(PwpVideo):
    def __init__(self, filename,  config: PwpConfig, working=None, backup=None):
        super().__init__(filename, config=config, working=working, backup=backup)

    def incr_logger_count(self, item=None):
        LOGGER.incr_picture("Mp4" + (f": {item}" if item else ""))

    def handles_metadata(self):
        return True


class PwpAvoided(PwpData):
    def __init__(self, filename):
        super().__init__(filename, None, None)
        self.to_abort = True

    def incr_logger_count(self, item=None):
        LOGGER.incr_picture("Not managed" + (f": {item}" if item else ""))


class PwpKept(PwpData):
    def __init__(self, filename):
        super().__init__(filename, None, None)

    def incr_logger_count(self, item=None):
        LOGGER.incr_picture("Kept" + (f": {item}" if item else ""))

    def handles_metadata(self):
        return False

    def patch_after_rename(self, new_filepath):
        """
        For KEPT objects, it is necessary to patch the name that rename generates:
        :param new_filepath: after rename
        :return: patched version
        """
        old_filename = os.path.basename(self.filename)
        new_filepath = os.path.dirname(new_filepath) + '/' + old_filename  # to keep filename
        return new_filepath



# ---------------------------------------------------------------------------------------------------------------
# piwiPre project
# This program and library is licenced under the European Union Public Licence v1.2 (see LICENCE)
# developed by fabien.battini(at)gmail.com
# ---------------------------------------------------------------------------------------------------------------

# CAVEAT: executes ONLY on windows !

import platform
import argparse
import termcolor
import requests
import shutil
import re
import sys
import os
import datetime

sys.path.append(os.path.dirname(os.path.dirname(os.path.realpath(__file__))))
from piwiPre.pwpVersion import PwpVersion
from piwiPre.pwpLogoSmall import pwpLogo_png

import winreg
import getpass
import zipfile
import json
import subprocess

import tkinter
from tkinter import ttk
from tkinter import scrolledtext
import tkinter.font

import ctypes
from ctypes import POINTER, c_ulong, c_char_p, c_int, c_void_p
from ctypes.wintypes import HANDLE, BOOL, DWORD, HWND, HINSTANCE, HKEY
from ctypes import windll

# Constant definitions

SEE_MASK_NOCLOSEPROCESS = 0x00000040        # noqa
SEE_MASK_NO_CONSOLE = 0x00008000

# Type definitions

PHANDLE = ctypes.POINTER(HANDLE)           # noqa
PDWORD = ctypes.POINTER(DWORD)             # noqa


class ShellExecuteInfo(ctypes.Structure):
    _fields_ = [
        ('cbSize',       DWORD),
        ('fMask',        c_ulong),
        ('hwnd',         HWND),
        ('lpVerb',       c_char_p),
        ('lpFile',       c_char_p),
        ('lpParameters', c_char_p),
        ('lpDirectory',  c_char_p),
        ('nShow',        c_int),
        ('hInstApp',     HINSTANCE),
        ('lpIDList',     c_void_p),
        ('lpClass',      c_char_p),
        ('hKeyClass',    HKEY),
        ('dwHotKey',     DWORD),
        ('hIcon',        HANDLE),
        ('hProcess',     HANDLE)]

    def __init__(self, **kw):
        super(ShellExecuteInfo, self).__init__()
        self.cbSize = ctypes.sizeof(self)
        for field_name, field_value in kw.items():
            setattr(self, field_name, field_value)


PShellExecuteInfo = POINTER(ShellExecuteInfo)


# Function definitions

ShellExecuteEx = windll.shell32.ShellExecuteExA
ShellExecuteEx.argtypes = (PShellExecuteInfo, )                  # noqa
ShellExecuteEx.restype = BOOL

WaitForSingleObject = windll.kernel32.WaitForSingleObject
WaitForSingleObject.argtypes = (HANDLE, DWORD)                   # noqa
WaitForSingleObject.restype = DWORD

CloseHandle = windll.kernel32.CloseHandle
CloseHandle.argtypes = (HANDLE, )                                # noqa
CloseHandle.restype = BOOL


# At last, the actual implementation!

# patch: execute the SAME python interpreter with the same cmdline arguments
#        so rely on orig_argv rather than argv
# patch: add --user username, to remember who was the user


def elevate(show_console=True):
    if windll.shell32.IsUserAnAdmin():
        return
    username = getpass.getuser()
    args = sys.orig_argv or sys.argv
    params = ShellExecuteInfo(
        fMask=SEE_MASK_NOCLOSEPROCESS | SEE_MASK_NO_CONSOLE,
        hwnd=None,
        lpVerb=b'runas',
        lpFile=args[0].encode('cp1252'),  # lpFile=sys.executable.encode('cp1252'),
        lpParameters=subprocess.list2cmdline(args[1:]+['--user', username]).encode('cp1252'),
        nShow=int(show_console))
    try:
        if not ShellExecuteEx(ctypes.byref(params)):
            raise ctypes.WinError()
    except OSError as e:
        print(f"OS Error {e}")
    handle = params.hProcess

    ret = DWORD()
    WaitForSingleObject(handle, -1)
    if windll.kernel32.GetExitCodeProcess(handle, ctypes.byref(ret)) == 0:
        raise ctypes.WinError()
    CloseHandle(handle)
    sys.exit(ret.value)


class PwpUi:
    def __init__(self, args, installer):
        self.installer = installer
        self.root = tkinter.Tk()
        self.root.title("piwiPre installer")
        self.frm = ttk.Frame(self.root, padding=10)
        self.frm.grid()
        self.args = args

        self.do_piwipre = tkinter.IntVar()
        self.do_mariadb = tkinter.IntVar()
        self.do_ffmpeg = tkinter.IntVar()
        self.do_mode = tkinter.StringVar()
        self.do_progress = tkinter.IntVar()
        self.do_msg = tkinter.StringVar()
        self.doc_url = "https://fabien_battini.gitlab.io/piwipre/html/"
        # self.piwipre_version = tkinter.StringVar()

        row = 0
        # CAVEAT: logo MUST be stored in an attribute, otherwise it is garbage collected !
        self.logo = pwpLogo_png.tk_photo()
        tkinter.Label(self.frm, image=self.logo).grid(column=0, row=row, sticky="W")

        title_font = tkinter.font.Font(size=15, family="Helvetica", weight="bold")
        lab = ttk.Label(self.frm, font=title_font,
                        text=f"------------- PiwiPre installer for Windows version {PwpVersion.spec} -------------\n")
        lab.grid(column=1, row=row, columnspan=5)

        row += 1
        ttk.Label(self.frm, text="Elevation enabled :", anchor="w", padding=4,
                  ).grid(column=0, row=row, columnspan=2, sticky="W")
        ttk.Label(self.frm, text=str(not self.installer.args.no_elevation), anchor="w",
                  ).grid(column=2, row=row, columnspan=2, sticky="W")

        row += 1
        ttk.Label(self.frm, text="Current Working Directory :", anchor="w", padding=4,
                  ).grid(column=0, row=row, columnspan=4, sticky="W")
        ttk.Label(self.frm, text=os.getcwd(), anchor="w",
                  ).grid(column=2, row=row, columnspan=2, sticky="W")

        row += 1
        ttk.Label(self.frm, text=f"Latest piwiPre version :", anchor="w", padding=4,
                  ).grid(column=0, row=row, columnspan=4, sticky="W")
        self.pwp_target_version = ttk.Label(self.frm, text=installer.piwipre_target_version)
        self.pwp_target_version.grid(column=2, row=row, columnspan=4, sticky="W")

        row += 1
        tkinter.Frame(self.frm, width=850, height=5, relief='sunken', background="#cccdddccc",
                      ).grid(column=0, row=row, columnspan=4)

        row += 1
        ttk.Label(self.frm, text=f"piwiPre EXE version in target dir:", anchor="w", padding=4,
                  ).grid(column=0, row=row, columnspan=4, sticky="W")
        self.pwp_version = ttk.Label(self.frm, text=installer.piwipre_version)
        self.pwp_version.grid(column=2, row=row, columnspan=4, sticky="W")

        row += 1
        ttk.Label(self.frm, text=f"ffmpeg installed version : ", anchor="w", padding=4,
                  ).grid(column=0, row=row, columnspan=4, sticky="W")
        self.ffmpeg_version = ttk.Label(self.frm, text=installer.ffmpeg_version, padding=4,)
        self.ffmpeg_version.grid(column=2, row=row, columnspan=4, sticky="W")

        row += 1
        ttk.Label(self.frm, text="piwiPre exe", anchor="w", padding=4,
                  ).grid(column=0, row=row, sticky="W")
        self.pwp_button = ttk.Checkbutton(self.frm, text="install", width=10,
                                          variable=self.do_piwipre)
        self.pwp_button.grid(column=1, row=row)

        self.pwp_path = ttk.Label(self.frm, text="",
                                  padding=4, width=90,  # background="light grey",
                                  anchor="w")
        self.pwp_path.grid(column=2, row=row)

        row += 1
        ttk.Label(self.frm, text="ffmpeg exe", anchor="w", padding=4,
                  ).grid(column=0, row=row, sticky="W")
        self.ffmpeg_button = ttk.Checkbutton(self.frm, text="install", width=10,
                                             variable=self.do_ffmpeg)
        self.ffmpeg_button.grid(column=1, row=row)

        self.ffmpeg_path = ttk.Label(self.frm, text="",
                                     padding=4,  width=90,  # background="light grey",
                                     anchor="w")
        self.ffmpeg_path.grid(column=2, row=row)

        row += 1
        ttk.Label(self.frm, text="MariaDB CC", anchor="w", padding=4,
                  ).grid(column=0, row=row, sticky="W")
        self.maria_button = ttk.Checkbutton(self.frm, text="install", width=10,
                                            variable=self.do_mariadb)
        self.maria_button.grid(column=1, row=row)

        row += 1
        tkinter.Frame(self.frm, width=850, height=10,
                      ).grid(column=0, row=row, columnspan=4)

        row += 1
        ttk.Radiobutton(self.frm, value="test", text="test", command=self.refresh_default_values,
                        variable=self.do_mode).grid(column=0, row=row, sticky="W")
        ttk.Radiobutton(self.frm, value="install", text="install", command=self.refresh_default_values,
                        variable=self.do_mode,).grid(column=1, row=row, sticky="W")

        tkinter.Button(self.frm, text="Run", command=self.run,
                       background="green", foreground="white",
                       activebackground="white", activeforeground="green"
                       ).grid(column=2, row=row, sticky="W")
        tkinter.Button(self.frm, text="  Quit  ", command=self.exit,
                       background="red", foreground="white",
                       activebackground="white", activeforeground="red").grid(column=3, row=row, sticky="W")

        row += 1
        tkinter.Frame(self.frm, width=850, height=10,
                      ).grid(column=0, row=row, columnspan=4)

        row += 1
        self.action = ttk.Label(self.frm, text="Downloader: idle", width=30, anchor="w", padding=4, )
        self.action.grid(column=0, row=row, columnspan=2, sticky="W")

        self.progress = ttk.Progressbar(self.frm, orient="horizontal", variable=self.do_progress,
                                        mode="determinate", length=550, maximum=100)
        self.progress.grid(column=2, row=row)
        self.action2 = tkinter.Label(self.frm, text="", width=30, anchor="w")
        self.action2.grid(column=3, row=row, columnspan=2, sticky="W")

        row += 1
        tkinter.Frame(self.frm, width=850, height=10,
                      ).grid(column=0, row=row, columnspan=4)
        row += 1
        ttk.Label(self.frm, text="Feedback from the installer",
                  padding=4,
                  anchor="w").grid(column=0, row=row, columnspan=5)

        row += 1
        font = tkinter.font.Font(size=10, family="Helvetica")
        self.msg = scrolledtext.ScrolledText(self.frm, background="#ccccccccc",
                                             padx=3, pady=3,
                                             font=font,
                                             width=100, height=10,
                                             )
        self.msg.grid(column=0, row=row, columnspan=5)
        self.msg.tag_config('warning', foreground="red")
        #
        row += 1
        tkinter.Frame(self.frm, width=850, height=5,
                      ).grid(column=0, row=row, columnspan=4)
        #
        row += 1
        ttk.Label(self.frm, text="Minimal help",
                  padding=4,
                  anchor="w").grid(column=0, row=row, columnspan=5)
        row += 1
        help_msg = """
        piwiPre Installer is available only for windows
        run piwiPreInstaller --help for the full help
        
        it can install:
        - piwiPre.exe, the application version of piwiPre that does NOT require python
          it will be installed in the system folders,
          and inserted in the PATH, so that you can run it from anywhere
        - ffmpeg.exe, the well-known tool used by piwiPre to manage VIDEO.
          If you do not need video, installing it is useless
        - MariaDB C Connector. It may be used is some cases to access to the remote SQL database of piwiGo.
          This is normally NOT required, unless piwiPre complains about its absence.
        
        In "test" mode, installation occurs in the current directory
        In "install" mode, installation proceeds normally
        
        """
        ttk.Label(self.frm, text=help_msg, width=100,).grid(column=0, row=row, columnspan=5)

        row += 1
        ttk.Label(self.frm, text="Documentation :", anchor="w",
                  ).grid(column=0, row=row, columnspan=2, sticky="W")

        ttk.Label(self.frm, text=self.doc_url, anchor="w",
                  ).grid(column=2, row=row, columnspan=2, sticky="W")
        tkinter.Button(self.frm, text="Copy URL", command=self.open_url,
                       activebackground="white", activeforeground="green"
                       ).grid(column=3, row=row, sticky="W")

        self.from_python_to_ui()

    def main_loop(self):
        self.root.mainloop()
        self.root.destroy()

    def refresh_default_values(self):
        self.from_ui_to_python()
        self.installer.compute_default_values()
        self.from_python_to_ui()

    def from_python_to_ui(self):
        self.installer.compute_default_values()
        self.do_piwipre.set(1 if self.installer.args.piwipre else 0)
        self.do_mariadb.set(1 if self.installer.args.mariadb else 0)
        self.do_ffmpeg.set(1 if self.installer.args.ffmpeg else 0)
        self.do_mode.set(self.installer.args.mode)

        self.pwp_button['text'] = "re install" if self.installer.piwipre_exists else "install"

        self.pwp_version['text'] = self.installer.piwipre_version or ''
        self.pwp_path['text'] = self.installer.piwipre_path

        if self.installer.ffmpeg_exists:
            self.ffmpeg_button['text'] = "re install"
        else:
            self.ffmpeg_button['text'] = "install"
        self.ffmpeg_version['text'] = self.installer.ffmpeg_version or ''
        self.ffmpeg_path['text'] = self.installer.ffmpeg_path
        self.do_progress.set(0)
        self.set_action("Downloader: idle", "")

    def from_ui_to_python(self):
        self.args.piwipre = self.do_piwipre.get() == 1
        self.args.ffmpeg = self.do_ffmpeg.get() == 1
        self.args.mariadb = self.do_mariadb.get() == 1
        self.args.mode = self.do_mode.get()
        pass

    def run(self):
        self.from_ui_to_python()
        self.installer.run()

    def add_msg(self, line, tag=None):
        if tag is None:
            self.msg.insert(tkinter.END, line)
        else:
            self.msg.insert(tkinter.END, line, tag)

    def exit(self):
        self.root.destroy()

    def set_action(self, line1, line2=""):
        self.action["text"] = line1
        self.action2["text"] = line2

    def set_progress(self, val):
        self.do_progress.set(int(val))
        self.root.update()
        self.root.update_idletasks()

    def open_url(self):
        self.frm.clipboard_clear()
        self.frm.clipboard_append(self.doc_url)
        self.frm.update()


class Installer:
    def __init__(self, arguments=None):
        self.progress_status = 0
        self.ui = None
        if platform.system() != "Windows":
            self.warning("--install-exe can be used only on Windows!")
            exit(-1)

        self.user_key = None
        arguments = arguments or []

        parser = argparse.ArgumentParser(description='install piwiPre on computer')
        parser.add_argument('--text-only',
                            help="does not display the graphical UI",
                            action='store_true')

        parser.add_argument('--user',
                            help="Install for user. This is for internal use only, not to be used by humans")
        parser.add_argument('--no-elevation',
                            help="Does not run privilege elevation, so install will probably fail",
                            action='store_true')
        parser.add_argument('--ffmpeg',
                            help="Install ffmpeg.exe",
                            action='store_true')
        parser.add_argument('--piwipre',
                            help="Install piwiPre.exe",
                            action='store_true')
        parser.add_argument('--mariadb',
                            help="Install mariaDb connector",
                            action='store_true')

        parser.add_argument('--force',
                            help="forces a new install of up to date packages (only for command-line)",
                            action='store_true')

        parser.add_argument('--mode',
                            help="test: run in separate dir, install: normal install",
                            action='store',
                            choices=['test', 'install'],
                            default="install")

        parser.add_argument('--chdir',
                            help="new directory to change to",
                            action='store')
        self.args = parser.parse_args(arguments)

        if self.args.chdir:
            self.makedirs(self.args.chdir)
            os.chdir(self.args.chdir)

        self.ffmpeg_path = None
        self.ffmpeg_exists = False
        self.ffmpeg_version = None

        self.piwipre_version = None
        self.piwipre_target_version = None
        self.piwipre_path = None
        self.piwipre_exists = False

        self.compute_default_values()

        if self.args.text_only:
            self.run()
        else:
            self.ui = PwpUi(self.args, installer=self)
            self.ui.main_loop()

    def compute_default_values(self):
        self.ffmpeg_version = None
        self.ffmpeg_path = os.getcwd() if self.args.mode == "test" else (os.environ['PROGRAMFILES(X86)'] + '\\ffmpeg')  # noqa

        self.ffmpeg_exists = (os.path.isfile(f"{self.ffmpeg_path}\\bin\\ffmpeg.exe") and
                              os.path.isfile(f"{self.ffmpeg_path}\\bin\\ffprobe.exe"))

        if self.ffmpeg_exists:
            self.action("Checking FFMPEG version", "ongoing")
            target = f"{self.ffmpeg_path}/bin/ffmpeg.exe"
            try:
                res = subprocess.run([target, "-version"], capture_output=True)
                buffer = res.stdout.decode('utf8')
                # ffmpeg version N-111004-g4893cbcaba-gb1c3d81e71+12 Copyright (c) 2000-2023 the FFmpeg developers  # noqa
                m = re.match(r".*ffmpeg version (.*) Copyright .*", buffer)
                if m:
                    self.ffmpeg_version = m.group(1)
            except subprocess.CalledProcessError as e:
                self.error(f"Error {e} while ffmpeg --version")

            if self.ui is None:
                self.msg(f"ffmpeg {self.ffmpeg_version} is already installed in '{self.ffmpeg_path}' ")
                if self.args.mode != "test":
                    self.msg("use --force to force new install")
                    self.args.ffmpeg = False
            self.action("", "")

        self.piwipre_path = os.getcwd() if self.args.mode == "test" else (os.environ['PROGRAMFILES(X86)'] + '\\piwiPre')  # noqa
        self.piwipre_exists = False
        self.piwipre_version = None

        if self.piwipre_target_version is None:
            pack = self.get_latest_package()
            self.piwipre_target_version = pack["version"]

        target = f"{self.piwipre_path}\\piwiPre.exe"

        if not os.path.isfile(target):
            return

        # HACK:
        # Do NOT ask me why!
        # when the exe is in "Program Files (x86)"
        # it is mandatory to have shell=True if we have stdout=subprocess.PIPE
        # otherwise run fails.
        # but, shell=True is NOT mandatory
        #   - if we do NOT have   stdout=subprocess.PIPE
        #   - OR if the exe is in a user directory
        # rather strange, and not so much documented.

        self.action("Checking PiwiPre installed version", "ongoing")
        try:
            res = subprocess.run([target, "--version", "--quiet"],
                                 stdout=subprocess.PIPE, shell=True,
                                 check=True, text=True)  # noqa
            buffer = res.stdout
            # current version: '0.17 at 03/30/2024 18:32:06'
            m = re.match(r"current version: '(.*)'", buffer)
            if m:
                self.piwipre_version = m.group(1)
        except OSError as e:
            self.warning(f"Error {e} while piwiPre --version")
            return False

        if self.piwipre_version is None:
            return
        self.action("", "")

        m = re.match(r"([\d.]*)", self.piwipre_version)
        self.piwipre_exists = m.group(1) >= self.piwipre_target_version

        if self.args.piwipre and self.ui is None:
            self.msg(f"piwiPre is already installed in '{self.piwipre_path}' " +
                     f" with version {self.piwipre_version}")
            if self.args.mode != "test":
                self.msg(f"use --force to force new install")
                self.args.piwipre = False

    def action(self, line1, line2=""):
        if self.ui:
            self.ui.set_action(line1, line2)

    def warning(self, line):
        if self.ui:
            self.ui.add_msg("WARNING: " + line + '\n', "warning")

        print(termcolor.colored("WARNING: " + line, color='red', force_color=True))

    def error(self, line):
        if self.ui:
            self.ui.add_msg("ERROR  : " + line + '\n', "warning")
        else:
            print(termcolor.colored("ERROR  : " + line, color='red', force_color=True))
            input("Close window after error?")
            exit(-1)

    def msg(self, line):
        if self.ui:
            self.ui.add_msg(line + "\n")
        print("msg     : " + line)

    @staticmethod
    def makedirs(path):
        if os.path.isdir(path):
            return
        os.makedirs(path)

    def get_html(self, url):
        response = requests.get(url)
        if response.status_code == 200:
            strings = response.content.decode('utf-8')
            return strings.split('\n')
        self.error(f"Failed to download '{url}'")

    def progress_bar(self, filename, nb_chunk: int, chunk_size: int, total: int):
        fetched_kb = int(nb_chunk * chunk_size / 1024)
        if nb_chunk <= 1:
            self.progress_status = 0

        if self.ui:
            pc = int(100 * 1024 * fetched_kb / total)
            self.action(filename, f"{fetched_kb: 6} KB / {int(total / 1024)} KB")
            self.ui.set_progress(pc)

        pc = int(50 * 1024 * fetched_kb / total)
        print(f"\r[{'#' * pc}{'-' * (49 - pc)}]  {fetched_kb: 6} KB / {int(total / 1024)} KB", end="")

    def download(self, url, dest):
        self.msg(f"Starting to download'{url}' into '{dest}'")
        response = requests.get(url, stream=True)
        size = int(response.headers['Content-Length'])
        self.msg(f"Size = {int(size / 1024)} KB")
        if response.status_code == 200:
            nb_chunk = 0
            self.makedirs(os.path.dirname(dest))
            with open(dest, "wb") as file:
                for chunk in response.iter_content(chunk_size=1024):
                    file.write(chunk)
                    nb_chunk += 1
                    self.progress_bar(os.path.basename(dest), nb_chunk, 1024, size)

            self.progress_bar("", 0, 1024, size)
            self.action(os.path.basename(dest), "downloaded")
            self.msg(f"\ndownloaded successfully '{url}' into '{dest}' in {int(size / 1024)} KB")
        else:
            self.error(f"Failed to download '{url}' into '{dest}'.")

    # -----------------------------------------------------------------------------------------
    # ffmpeg

    def install_ffmpeg(self):
        url = "https://www.gyan.dev/ffmpeg/builds/ffmpeg-release-essentials.zip"

        self.msg(f"Downloading latest FFMPEG package version {url}")
        self.download(url, f"{self.ffmpeg_path}/ffmpeg.zip")
        root = None
        with zipfile.ZipFile(f"{self.ffmpeg_path}/ffmpeg.zip") as myzip:
            names = myzip.namelist()
            found_ffmpeg = False
            found_ffprobe = False
            for f in names:
                if root is None or len(f) < len(root):
                    root = f
                if "ffmpeg.exe" in f:
                    myzip.extract(f, self.ffmpeg_path)
                    found_ffmpeg = True
                    self.makedirs(f"{self.ffmpeg_path}/bin")
                    shutil.move(f"{self.ffmpeg_path}/{f}", f"{self.ffmpeg_path}/bin/ffmpeg.exe")
                if "ffprobe.exe" in f:
                    myzip.extract(f, self.ffmpeg_path)
                    found_ffprobe = True
                    self.makedirs(f"{self.ffmpeg_path}/bin")
                    shutil.move(f"{self.ffmpeg_path}/{f}", f"{self.ffmpeg_path}/bin/ffprobe.exe")

        os.remove(f"{self.ffmpeg_path}/ffmpeg.zip")
        shutil.rmtree(f"{self.ffmpeg_path}/{root}")
        if found_ffmpeg and found_ffprobe:
            self.msg(f"ffmpeg installed in '{self.ffmpeg_path}'")
        else:
            self.error('ffmpeg or ffprobe not found in archive')

        # useless: ffmpeg is called from absolute path
        # if not self.args.self_test:
        #     self.add_to_path(ffmpeg_path + '/bin')

    # -----------------------------------------------------------------------------------------
    # auto-install

    def get_json(self, url):
        response = requests.get(url)
        if response.status_code == 200:
            res = json.loads(response.content)
            return res
        self.error(f"Failed to download json from '{url}'")

    def get_latest_package(self):
        """returns the latest package"""
        latest = None
        packages = self.get_json("https://gitlab.com/api/v4/projects/48915444/packages")
        for pack in packages:
            if latest is None:
                latest = pack
            elif pack['version'] > latest['version']:
                latest = pack

        return latest

    def install_piwipre(self):
        self.makedirs(self.piwipre_path)
        target = self.piwipre_target_version

        base = "https://gitlab.com/api/v4/projects/48915444/packages/generic/piwiPre"
        self.download(f"{base}/{target}/piwiPre-{target}.exe",
                      f"{self.piwipre_path}/piwiPre.exe")
        shutil.copy2(f"{self.piwipre_path}/piwiPre.exe",
                     f"{self.piwipre_path}/piwiPre-{target}.exe")

        self.warning(f"Latest package version '{target}' installed in {self.piwipre_path}")

        if not self.args.mode == "test":
            self.add_to_path(os.environ['PROGRAMFILES(X86)'] + '\\piwiPre')  # noqa

    def find_maria_db_url(self, bits: str):
        all_versions = self.get_html("https://dlm.mariadb.com/browse/c_connector/")
        url = None
        version = None
        for line in all_versions:
            #             <td><a href="/browse/c_connector/201/">C connector 3.3</a></td>
            #             <td><a href="/browse/c_connector/169/">C connector 3.2 (EOL)</a></td>
            m = re.match(r'\s*<td><a href="(.*)">C connector (.*)</a></td>', line)
            if m and m.group(2) and 'EOL' not in m.group(2):
                if version is None or m.group(2) > version:
                    version = m.group(2)
                    url = m.group(1)
        if url is None:
            self.error('Unable to find URL of current version in "https://dlm.mariadb.com/browse/c_connector/"')

        all_sub_versions = self.get_html("https://dlm.mariadb.com" + url)
        sub_url = None
        sub_version = None
        for line in all_sub_versions:
            #             <td><a href="/browse/c_connector/201/1294/">C connector 3.3.0</a></td>
            m = re.match(r'\s*<td><a href="(.*)">C connector (.*)</a></td>', line)
            if m and 'EOL' not in m.group(2):
                if sub_version is None or m.group(2) > sub_version:
                    sub_version = m.group(2)
                    sub_url = m.group(1)

        if sub_url is None:
            self.error(f'Unable to find current version in "https://dlm.mariadb.com/browse/c_connector{url}"')

        all_bins = self.get_html("https://dlm.mariadb.com" + sub_url)
        for line in all_bins:
            # <td><a href="https://dlm.mariadb.com/3677107/Connectors/c/connector-c-3.3.8/
            # mariadb-connector-c-3.3.8-win32.msi">connector-c-3.3.8/mariadb-connector-c-3.3.8-win32.msi</a></td>
            m = re.match(r'\s*<td><a href="(.*)">.*\.msi</a></td>', line)
            if m and bits in m.group(1):
                return m.group(1)
        self.error(f'Unable to {bits} in "https://dlm.mariadb.com{sub_url}"')

    def install_maria_db(self):
        archi = "64" if "64" in os.environ['PROCESSOR_ARCHITECTURE'] else '32'
        url = self.find_maria_db_url(archi)
        dest = f"{os.getcwd()}/mariadb.msi"
        self.download(url, dest)
        self.warning(f"maria_db for {archi} bits architecture is downloaded as '{dest}'")

        if self.args.mode == "test":
            self.warning("You should running it AS AN ADMINISTRATOR")
            #  os.system(dest)
        else:
            self.warning("test: NOT running it ")

    def install_for_python(self):
        """
        Used in a PYTHON context, for windows architectures
        helps the installation of ffmpeg and mariaDb
        :return:
        """
        if platform.system() != "Windows":
            self.error("--install-tools can be used only on Windows!")

        self.install_ffmpeg()
        self.install_maria_db()

    @staticmethod
    def get_user_from_key(key):
        try:
            with winreg.ConnectRegistry(None, winreg.HKEY_USERS) as registry:
                with winreg.OpenKey(registry, key) as key2:
                    with winreg.OpenKey(key2, "Volatile Environment") as key3:
                        return winreg.QueryValueEx(key3, "USERNAME")[0]
        except OSError:
            return None

    @staticmethod
    def get_value(key, fid):
        try:
            return winreg.QueryValueEx(key, fid)[0]
        except OSError:
            return None

    def get_user_key(self, username):
        # How to find user registry ID?
        # look at all HKEY-USERS/key/Volatile Environmemt/USERNAME      # noqa
        with winreg.ConnectRegistry(None, winreg.HKEY_USERS) as registry:
            index = 0
            while True:
                try:
                    key = winreg.EnumKey(registry, index)
                    val = self.get_user_from_key(key)
                    if val == username:
                        return key
                    index += 1
                except OSError:
                    return None

    def add_to_path(self, value, master_key="Path"):
        if not windll.shell32.IsUserAnAdmin():
            self.error("INTERNAL ERROR: Not an admin")

        try:
            with winreg.ConnectRegistry(None, winreg.HKEY_USERS) as registry:
                with winreg.OpenKey(registry, self.user_key) as key2:
                    with winreg.CreateKey(key2, "Environment") as key3:  #
                        current = self.get_value(key3, master_key)
                        if current is None:
                            current = value + ";"
                            winreg.SetValueEx(key3, master_key, 0, winreg.REG_EXPAND_SZ, current)
                            winreg.FlushKey(key3)
                            self.msg(f"Created '{value}' to {master_key} environment variable")
                        elif value not in current:
                            if current[-1:] != ';':
                                current += ';'
                            current += value + ";"
                            winreg.SetValueEx(key3, master_key, 0, winreg.REG_EXPAND_SZ, current)
                            winreg.FlushKey(key3)
                            self.msg(f"Added '{value}' to {master_key} environment variable")
                        else:
                            self.msg(f"{master_key} already get '{value}' ")
        except Exception as e:
            self.error(f"Exception '{e}' ")

    def run_min_test(self):
        self.action("Minimal test", "Starting")
        test_dir = os.environ['PROGRAMFILES(X86)'] + '\\piwiPreTmp'   # noqa
        if os.path.isdir(test_dir):
            os.rmdir(test_dir)
            self.msg(f"Removed {test_dir}")
        else:
            os.makedirs(test_dir)
            self.msg(f"Created {test_dir}")
        self.add_to_path(str(datetime.datetime.now()), "PiwiPrePath")
        self.action("Minimal test", "Done")

    def run(self):
        self.action("Installation", "Starting")

        if self.args.user:
            self.user_key = self.get_user_key(self.args.user)
        elif self.args.no_elevation and self.args.mode == "test":
            pass   # this is normal
        else:
            self.error("user should have been passed on cmdline")

        if not self.args.piwipre and not self.args.ffmpeg and not self.args.mariadb:
            self.run_min_test()
        if self.args.piwipre:
            self.install_piwipre()
        if self.args.ffmpeg:
            self.install_ffmpeg()
        if self.args.mariadb:
            self.install_maria_db()
        self.action("Installation", "Done")
        if self.ui is None and self.args.mode == "install":
            # if no-elevation,
            input("Close window?")


def run_installer(arguments):
    if platform.system() != "Windows":
        print("runs only on windows")
        return
    if "--no-elevation" not in sys.argv:
        elevate(show_console="--text-only" in sys.argv)   # noqa
    try:
        Installer(arguments)
    except OSError as e:
        print(f"OS Error {e}")


if __name__ == "__main__":
    run_installer(sys.argv[1:])
